"""
This module contains functions for augmenting images that are
suited to remote sensing imagery.
"""
# Standard library
import sys; sys.path.append("../../")
from typing import Optional, Tuple, List

# External
import numpy as np

# Internal
from buteo.ai.augmentation_batch import (
    augmentation_batch_rotation,
    augmentation_batch_mirror,
    augmentation_batch_channel_scale,
    augmentation_batch_noise,
    augmentation_batch_contrast,
    augmentation_batch_blur,
    augmentation_batch_sharpen,
    augmentation_batch_drop_pixel,
    augmentation_batch_drop_channel,
    augmentation_batch_misalign,
    augmentation_batch_cutmix,
    augmentation_batch_mixup,
)


def apply_augmentations(
    batch_x: np.ndarray,
    batch_y: Optional[np.ndarray],
    *,
    augmentations: Optional[List[dict]] = None,
    channel_last: bool = True,
) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """
    Apply a list of augmentations to a batch of images.
    
    Args:
        batch_x (np.ndarray): The batch of images to augment.
        batch_y (np.ndarray/None=None): The batch of labels to augment.

    Keyword Args:
        augmentations (list/dict=None): The list of augmentations to apply.
        channel_last (bool=True): Whether the image is (channels, height, width) or (height, width, channels).

    Returns:
        Tuple(np.ndarray, np.ndarray): The augmented batch of images and labels (if provided).
    """

    if augmentations is None or len(augmentations) == 0:
        return batch_x, batch_y

    batch_x, batch_y = batch_x.copy(), batch_y.copy()

    for aug in augmentations:
        if aug["name"] == "rotation":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_rotation(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "mirror":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_mirror(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "channel_scale":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "amount" in aug:
                kwargs["amount"] = aug["amount"]
            if "additive" in aug:
                kwargs["additive"] = aug["additive"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_channel_scale(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "noise":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "amount" in aug:
                kwargs["amount"] = aug["amount"]
            if "additive" in aug:
                kwargs["additive"] = aug["additive"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_noise(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "contrast":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "contrast_factor" in aug:
                kwargs["contrast_factor"] = aug["contrast_factor"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_contrast(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "drop_pixel":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "drop_probability" in aug:
                kwargs["drop_probability"] = aug["drop_probability"]
            if "drop_value" in aug:
                kwargs["drop_value"] = aug["drop_value"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_drop_pixel(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "drop_channel":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "drop_probability" in aug:
                kwargs["drop_probability"] = aug["drop_probability"]
            if "drop_value" in aug:
                kwargs["drop_value"] = aug["drop_value"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_drop_channel(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "blur":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "intensity" in aug:
                kwargs["intensity"] = aug["intensity"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_blur(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "sharpen":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "intensity" in aug:
                kwargs["intensity"] = aug["intensity"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_sharpen(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "misalign":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "max_offset" in aug:
                kwargs["max_offset"] = aug["max_offset"]
            if "max_images" in aug:
                kwargs["max_images"] = aug["max_images"]
            if "max_channels" in aug:
                kwargs["max_channels"] = aug["max_channels"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]

            batch_x, batch_y = augmentation_batch_misalign(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "cutmix":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "max_size" in aug:
                kwargs["max_size"] = aug["max_size"]
            if "max_mixes" in aug:
                kwargs["max_mixes"] = aug["max_mixes"]
            if "feather" in aug:
                kwargs["feather"] = aug["feather"]
            if "feather_dist" in aug:
                kwargs["feather_dist"] = aug["feather_dist"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_cutmix(
                batch_x,
                batch_y,
                **kwargs,
            )

        elif aug["name"] == "mixup":
            kwargs = {}
            if "chance" in aug:
                kwargs["chance"] = aug["chance"]
            if "max_mixes" in aug:
                kwargs["max_mixes"] = aug["max_mixes"]
            if "channel_last" in aug:
                kwargs["channel_last"] = aug["channel_last"]
            else:
                kwargs["channel_last"] = channel_last

            batch_x, batch_y = augmentation_batch_mixup(
                batch_x,
                batch_y,
                **kwargs,
            )

    return batch_x, batch_y


def augmentation_generator(
    X: np.ndarray,
    y: Optional[np.ndarray],
    batch_size: int = 64,
    augmentations: Optional[List[dict]] = None,
    shuffle: bool = True,
    seed: Optional[int] = None,
    channel_last: bool = True,
) -> np.ndarray:
    """
    Generate batches of augmented data.
    
    Args:
        X (np.ndarray): The data to augment.

    Keyword Args:
        y (np.ndarray): The labels for the data.
        batch_size (int): The size of the batches to generate.
        augmentations (list): The augmentations to apply.
        shuffle (bool): Whether to shuffle the data before generating batches.
        seed (int): The seed to use for shuffling.
        channel_last (bool): Whether the data is in channel last format.

    Returns:
        A generator yielding batches of augmented data.
    """
    if seed is not None:
        np.random.seed(seed)

    if shuffle:
        idx = np.random.permutation(X.shape[0])
        X = X[idx]
        y = y[idx]

    num_batches = (X.shape[0] + batch_size - 1) // batch_size

    for i in range(num_batches):
        start_idx = i * batch_size
        end_idx = min((i + 1) * batch_size, X.shape[0])

        batch_x = X[start_idx:end_idx]
        batch_y = y[start_idx:end_idx]

        batch_augmented_x, batch_augmented_y = apply_augmentations(
            batch_x,
            batch_y,
            augmentations=augmentations,
            channel_last=channel_last,
        )

        yield batch_augmented_x, batch_augmented_y


class AugmentationDataset:
    """
    A dataset that applies augmentations to the data.

    Args:
        X (np.ndarray): The data to augment.

    Keyword Args:
        y (np.ndarray): The labels for the data.
        batch_size (int): The size of the batches to generate.
        augmentations (list): The augmentations to apply.
        shuffle (bool): Whether to shuffle the data before generating batches.
            and whenever the dataset is iterated over.
        seed (int): The seed to use for shuffling.
        channel_last (bool): Whether the data is in channel last format.
    
    Returns:
        A dataset yielding batches of augmented data. For Pytorch,
            convert the batches to tensors before ingestion.
    """
    def __init__(
        self,
        X: np.ndarray,
        *,
        y: Optional[np.ndarray],
        batch_size: int = 64,
        augmentations: Optional[List[dict]] = None,
        shuffle: bool = True,
        seed: Optional[int] = None,
        channel_last: bool = True,
    ):
        self.X = X
        self.y = y
        self.batch_size = batch_size
        self.augmentations = augmentations
        self.shuffle = shuffle
        self.seed = seed
        self.channel_last = channel_last
        self.generator = None

    def __iter__(self):
        self.generator = augmentation_generator(
            self.X,
            y=self.y,
            batch_size=self.batch_size,
            augmentations=self.augmentations,
            shuffle=self.shuffle,
            seed=self.seed,
            channel_last=self.channel_last,
        )
        return self

    def __next__(self):
        try:
            return next(self.generator)
        except StopIteration as e:
            raise StopIteration from e

    def __len__(self):
        num_batches = (self.X.shape[0] + self.batch_size - 1) // self.batch_size

        return num_batches



"""
This module provides a set of functions to normalise data for machine learning.
"""

# Standard library
from typing import Union

# External
import numpy as np


def scaler_minmax(arr: np.array) -> np.array:
    """ 
    Normalise an array using the min-max method.

    Args:
        arr (np.array): The array to normalise.

    Returns:
        np.array: The normalised array. (dtype: float32)
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."

    num = np.subtract(arr, arr.min())
    den = np.subtract(arr.max(), arr.min())

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    return result


def scaler_standardise(arr: np.array) -> np.array:
    """ 
    Normalise an array using the mean and standard deviation method.

    Args:
        arr (np.array): The array to normalise.

    Returns:
        np.array: The normalised array. (dtype: float32)
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."

    mean = np.mean(arr)
    std = np.std(arr)

    result = np.zeros_like(arr, dtype="float32")

    num = np.subtract(arr, mean)
    np.divide(num, std, out=result, where=std != 0)

    return result


def scaler_standardise_mad(arr: np.array) -> np.array:
    """ Normalise an array using the median absolute deviation (MAD) method.
    
    Args:
        arr (np.array): The array to normalise.

    Returns:
        np.array: The normalised array. (dtype: float32)
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."

    median = np.median(arr)
    absdev = np.abs(np.subtract(arr, median))
    madstd = np.median(absdev) * 1.4826

    result = np.zeros_like(arr, dtype="float32")

    sub = np.subtract(arr, median)
    np.divide(sub, madstd, out=result, where=madstd != 0)

    return result


def scaler_iqr(
    arr: np.array,
    q1=0.25,
    q3=0.75,
) -> np.array:
    """ 
    Normalise an array using the interquartile range (IQR) method.

    Args:
        arr (np.array): The array to normalise.

    Keyword Args:
        q1 (float=0.25): The lower quartile.
        q3 (float=0.75): The upper quartile.

    Returns:
        np.array: The normalised array. (dtype: float32)
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."
    assert q1 < 0.5 and q1 >= 0.0, "q1 must be less than 0.5 and above or equal to 0.0"
    assert q3 > 0.5 and q3 <= 1.0, "q3 must be greater than 0.5 and below or equal to 1.0"
    assert q1 < q3, "q1 must be less than q3"

    q1, median, q3 = np.quantile(arr, [q1, 0.5, q3])

    num = np.subtract(arr, median)
    den = np.subtract(q3, q1)

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    return result


def scaler_to_range(
    arr: np.ndarray,
    min_val: Union[float, int] = 0.0,
    max_val: Union[float, int] = 1.0,
):
    """ 
    Normalise an array to a given range.

    Args:
        arr (np.array): The array to normalise.

    Keyword Args:
        min_val (float=0.0): The minimum value.
        max_val (float=1.0): The maximum value.

    Returns:
        np.array: The normalised array. (dtype: float32)
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."

    num = np.subtract(arr, arr.min())
    den = np.subtract(arr.max(), arr.min())

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    return np.multiply(result, max_val - min_val) + min_val


def scaler_truncate(
    arr: np.ndarray,
    trunc_min: Union[float, int],
    trunc_max: Union[float, int],
    target_min: Union[float, int] = 0.0,
    target_max: Union[float, int] = 1.0,
) -> np.ndarray:
    """
    Truncate an array and then normalise it to a given range.

    Args:
        arr (np.array): The array to normalise
        trunc_min (float/int): The minimum value to truncate to.
        trunc_max (float/int): The maximum value to truncate to.

    Keyword Args:
        target_min (float=0.0): The minimum value to normalise to.
        target_max (float=1.0): The maximum value to normalise to.
    
    Returns:
        np.array: The normalised array. (dtype: float32)
    """
    assert isinstance(arr, np.ndarray), "Input must be a numpy array."

    truncated = np.where(arr > trunc_max, trunc_max, arr)
    truncated = np.where(truncated < trunc_min, trunc_min, truncated)

    num = np.subtract(truncated, truncated.min())
    den = np.subtract(truncated.max(), truncated.min())

    result = np.zeros_like(arr, dtype="float32")
    np.divide(num, den, out=result, where=den != 0)

    return np.multiply(result, target_max - target_min) + target_min
